--****************************************************************************** 
--*COPYRIGHT: BENJAMIN GRYDEHOEJ - WWW.BG-ELEKTRONIK.DK - 2006 - FPGA SPARTAN-3* 
--******************************************************************************
--  Author:          Benjamin Grydehoej
--  Create the:      11th November, 2005
--  Last update the: 11th April, 2006
--  File:            IO_PORTS.VHD
--  Target Hardware: Xilinx Spartan3 - XC3S200
--  Tool chain:      Xilinx - Project Navicator 6.3.03i
--  Version:         1.0.C
--
--  DESCRIPTION:
--  ============
--  This VHDL code controls the Input/Output Ports, Serial UART and the SFR
--  register with Interrupt controls and Timer function. The Sub-functions 
--  is described in the code before the program. All commandos for 
--  communication in this program are control by the Special Function Register
--  there is listed under this text.
--  ======================================================
--   		SPECIAL FUNCTION REGISTER:
--  ======================================================
--   Symbol:	Name:						Address:
--	P0		Port 0						HEX 01
--	P1		Port 1						HEX 02
--	P2L		Port 2 Address Bus low byte 		HEX 03
--	P2H		Port 2 Address Bus High byte		HEX 04
--	P3		Port 3						HEX 05
--	P4		Port 4						HEX 06
--	P5		Port 5						HEX 07
--	SBUF	  	Serial channel buffer register	HEX 08
--	TLBS		Timer Low BAUE Rate Serial		HEX 09
--	THBS		Timer HIGH BAUE Rate Serial		HEX 0A
--	SCON		Serial channel control register	HEX 0B
--	IEN0		Interrupt enable register 0		HEX 0C
--	IEN1		Interrupt enable register 1		HEX 0D
--	ISC0		Interrupt service control register	HEX 0E
--	TCON		Timer service control register	HEX 0F
--	TC0		Timer Count 0 					HEX 10
--	TCL1		Timer Count Low 1 				HEX 11
--	TCH1		Timer Count High 1 				HEX 12
--------------------------------------------------------------------------
-- Standard IEEE libraries
--------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
--------------------------------------------------------------------------
entity in_out_ports is
  Port ( 	          CLK_io : in  	std_logic;
			   Reset_io : in  	std_logic;
    		           WE_io : in  	std_logic;
			      RE_io : in  	std_logic;
   	                ID_io : in  	std_logic_vector(7 downto 0);
			 Data_in_io : in  	std_logic_vector(7 downto 0);
			Data_out_io : out 	std_logic_vector(7 downto 0);
		    Interrupt_io : out 	std_logic;
		Interrupt_ack_io : in 	std_logic;
		 rx_data_present : in  	std_logic;
    			  Port_0_io : inout std_logic_vector(7 downto 0);
  			  Port_1_io : inout std_logic_vector(7 downto 0);
  			  Port_2_io : out 	std_logic_vector(15 downto 0);
  			  Port_3_io : inout std_logic_vector(7 downto 0);
			  Port_4_io : inout std_logic_vector(7 downto 0);
			  Port_5_io : inout std_logic_vector(7 downto 0);
			   TBS_uart : out  	std_logic_vector(15 downto 0);
		        BDP_uart : in  	std_logic;
			   RBH_uart : in  	std_logic;
			   RBF_uart : in  	std_logic;
			   TBH_uart : in  	std_logic;
			   TBF_uart : in  	std_logic;
		   write_to_uart : out 	std_logic;
		    data_to_uart : out 	std_logic_vector(7 downto 0);
		  read_from_uart : out 	std_logic;
		  data_from_uart : in  	std_logic_vector(7 downto 0);
			  TR0_timer : inout std_logic;
			  TF0_timer : in  	std_logic;
			  TC0_timer : out 	std_logic_vector(7 downto 0);
			  TR1_timer : inout std_logic;
			  TF1_timer : in  	std_logic;
			  TC1_timer : out 	std_logic_vector(15 downto 0));	
		  
end in_out_ports;

--------------------------------------------------------------------------
--  This program function is the handling of Input and Output control 
--  after the Special Function Register (SFR) for read and write to 
--  ports and option of Interrupt Service Routines and Timer control etc.
-------------------------------------------------------------------------- 
architecture in_out_ports of in_out_ports is

--------------------------------------------------------------------------
-- Signals for Interrupt connections
--------------------------------------------------------------------------
-- IEN0 - Interrupt Enables 0 Variables
signal     EA_int : std_logic;
signal    ET1_int : std_logic;
signal    ET0_int : std_logic;
signal    EX2_int : std_logic; 
signal    EX1_int : std_logic; 
signal    EX0_int : std_logic;
Signal	ES0_int : std_logic;
signal    IC0_int : std_logic;
signal    IS0_int : std_logic;
signal    IT2_int : std_logic;
signal    IT1_int : std_logic;
signal    IT0_int : std_logic;
signal    IX2_int : std_logic;
signal    IX1_int : std_logic;
signal    IX0_int : std_logic;
-- Interrupt Variables
signal     X0_int : std_logic;
signal	 X1_int : std_logic;
signal	 X2_int : std_logic;

begin
   process
   begin
 
      wait until (CLK_io'event and CLK_io='1');
	 X0_int <= Port_3_io(0); -- Read the Port 1 bit 0 value and save 
	 					-- it in the Internal signal. 
	 X1_int <= Port_3_io(1);	-- Read the Port 1 bit 1 value and save 
	 					-- it in the Internal signal. 
	 X2_int <= Port_3_io(2); -- Read the Port 1 bit 2 value and save 
	 					-- it in the Internal signal.
-------------------------------------------------------------------------
--	 RESET:
--    This function set ports level after reset and 
--	 define the value for variables
-------------------------------------------------------------------------
      if(Reset_io='1') then
	    -- PORTS  level after reset
	    Port_0_io <= "ZZZZZZZZ";  -- Set Port 0 to high impedance level
  	    Port_1_io <= "ZZZZZZZZ";  -- Set Port 1 to high impedance level
 	    Port_2_io <= X"0000";     -- Set Port 2 to Address 0 (Hexadecimal)
  	    Port_3_io <= "ZZZZZZZZ";  -- Set Port 3 to high impedance level
	    Port_4_io <= "ZZZZZZZZ";  -- Set Port 4 to high impedance level
	    Port_5_io <= "ZZZZZZZZ";  -- Set Port 5 to high impedance level
	    -- TIMER
	    TBS_uart  <= X"0146";	-- Set Baud Rate at 9600 (dec 326) HEX 0146
	    TC0_timer <= X"FF";		-- Set Timer Value to 255
	    TC1_timer <= X"FFFF";	-- Set Timer Value to 65535
	    TR0_timer <= '0';		-- Clear Timer 0 Run (Timer stop)
	    TR1_timer <= '0';		-- Clear Timer 1 Run (Timer stop)
	    -- INTERRUPT SYSTEM
	    Interrupt_io <= '0'; 	-- Set global Interrupt to zero
	    -- IEN0 - Interrupt Enables 0
	    EA_int	<= '0';		-- Clear Enable All Interrupts
--	    WDT_int	<= '0';		-- Clear WDT Interrupt 
--	    ET2_int	<= '0';		-- Clear Enable Timer 2 Interrupt
	    ET1_int	<= '0';		-- Clear Enable Timer 1 Interrupt
	    ET0_int	<= '0';		-- Clear Enable Timer 0 Interrupt
	    EX2_int	<= '0';		-- Clear Enable Extern 2 Interrupt
	    EX1_int	<= '0';		-- Clear Enable Extern 1 Interrupt
	    EX0_int	<= '0';		-- Clear Enable Extern 0 Interrupt
	    -- IEN1 - Interrupt Enables 1
--	    EC0_int	<= '0';		-- Clear Enable CAN Interrupt
	    ES0_int	<= '0';		-- Clear Enable Serial Interrupt
	    -- ISC0 - Interrupt Service Control
         IC0_int 	<= '0';		-- Clear FLAG CAN-BUS Interrupt
	    IS0_int 	<= '0';		-- Clear FLAG Serial Interrupt
	    IT2_int 	<= '0';		-- Clear FLAG Timer Overflow 2 Interrupt
	    IT1_int 	<= '0';		-- Clear FLAG Timer Overflow 1 Interrupt
	    IT0_int 	<= '0';		-- Clear FLAG Timer Overflow 0 Interrupt
	    IX2_int 	<= '0';		-- Clear FLAG Extern 2 Interrupt
	    IX1_int 	<= '0';		-- Clear FLAG Extern 1 Interrupt
	    IX0_int 	<= '0';		-- Clear FLAG Extern 0 Interrupt

--------------------------------------------------------------------------
--   Input & Output Interface:
--	This program function looks at the incoming port address from the 
--	PicoBlaze processor core and control via write enable (WE_io) and 
--	the read enable (RE_io) for Read or write to the Ports the ID is 
--	control from the processor and set the SFR value to the right port.    
--------------------------------------------------------------------------
--	 Port 0 I/O - SFR P0, HEX 01 DataBus
	 elsif(WE_io = '1'  and ID_io = X"01") then
	    Port_0_io <= Data_in_io; -- Send data from Microcontroller to Port 0
      elsif(RE_io  = '1' and ID_io = X"01") then
         Data_out_io <= Port_0_io;-- Send data from Port 0 to Microcontroller
--	 Port 1 I/O - SFR P1, HEX 02
	 elsif(WE_io = '1'  and ID_io = X"02") then
	    Port_1_io <= Data_in_io; -- Send data from Microcontroller to Port 1
      elsif(RE_io  = '1' and ID_io = X"02") then
         Data_out_io <= Port_1_io;-- Send data from Port 1 to Microcontroller
--	 Port 2 I/O - SFR P2, HEX 03 Address Bus Low and HEX 04 Address Bus High 
	 elsif(WE_io = '1'  and ID_io = X"03") then
	    Port_2_io(7 DOWNTO 0) <= Data_in_io(7 DOWNTO 0); -- Send low byte to Port 2
      elsif(WE_io = '1'  and ID_io = X"04") then
         Port_2_io(15 DOWNTO 8) <= Data_in_io(7 DOWNTO 0);-- Send high byte to Port 2
--	 Port 3 I/O - SFR P3, HEX 05
	 elsif(WE_io = '1'  and ID_io = X"05") then
	    Port_3_io <= Data_in_io; -- Send data from Microcontroller to Port 3
      elsif(RE_io  = '1' and ID_io = X"05") then
         Data_out_io <= Port_3_io;-- Send data from Port 3 to Microcontroller
--	 Port 4 I/O - SFR P4, HEX 06
	 elsif(WE_io = '1'  and ID_io = X"06") then
	    Port_4_io <= Data_in_io; -- Send data from Microcontroller to Port 4
      elsif(RE_io  = '1' and ID_io = X"06") then
         Data_out_io <= Port_4_io;-- Send data from Port 4 to Microcontroller
--	 Port 5 I/O - SFR P5, HEX 07
	 elsif(WE_io = '1'  and ID_io = X"07") then
	    Port_5_io <= Data_in_io; -- Send data from Microcontroller to Port 5
      elsif(RE_io  = '1' and ID_io = X"07") then
         Data_out_io <= Port_5_io;-- Send data from Port 5 to Microcontroller
--------------------------------------------------------------------------
--  SERIAL DATA CONTROL:
--  This function controls the Serial UART with write and read via SBUF and 
--  the option for the BAUD rate timing there control the speed for the serial
--  communication. The Serial Channel Control register SCON do it possible to 
--  read the status flag for the received and transmit buffer and look after 
--  the status flag BDP for receive data.  
--------------------------------------------------------------------------
--	 SBUF Read and write to Comport - SFR SBUF, HEX 08
	 elsif(WE_io = '1'  and ID_io = X"08") then -- Write to Serial Buffer
	    write_to_uart <= WE_io;		-- Enable write to UART
	    data_to_uart  <= Data_in_io;	-- Send Data to the UART buffer from
	    							-- the SFR register named SBUF
      elsif(RE_io  = '1' and ID_io = X"08") then -- Read to Serial Buffer
	    read_from_uart <= RE_io;		-- Enable read to UART 
         Data_out_io <= data_from_uart;	-- Read Data from the UART buffer to
								-- the SFR register named SBUF
--	 Timer Baud rate serial, low byte - SFR TLBS, HEX 09
	 elsif(WE_io = '1'  and ID_io = X"09") then -- Write the low byte
	    TBS_uart(7 DOWNTO 0) <= Data_in_io; -- Send the low data byte to the Timer  
	    								-- Baud rate Serial for BAUD rate timing
--	 Timer Baud rate serial, high byte - SFR THBS, HEX 0A
      elsif(WE_io  = '1' and ID_io = X"0A") then -- Write the high byte
         TBS_uart(15 DOWNTO 8) <= Data_in_io;-- Send the high data byte to the Timer  
	    								-- Baud rate Serial for BAUD rate timing

--	 Serial Channel Control Register - SFR SCON, HEX 0B
--     -----------------------------------------------
--    |  X  |  X  |	 X  | BDP | RBH | RBF | TBH | TBF |
--	  -----------------------------------------------
      elsif(RE_io  = '1' and ID_io = X"0B") then  -- Read the status flag from Serial 
	 									-- Channel Control Register SCON
         Data_out_io <= ("000" & BDP_uart & RBH_uart & RBF_uart & TBH_uart & TBF_uart);

-- Serial Interrupt handling 		
 	 elsif (ES0_int='1' and IS0_int='0' and rx_data_present='1' and EA_int='1') then
	   IS0_int <= '1'; -- Set Serial Interrupt FLAG

--------------------------------------------------------------------------	    
--  INTERRUPT SYSTEM:
--  This program handle the Interrupt System there use the three register 
--  named IEN0 - Interrupt Enable 0, IEN1 - Interrupt Enable 2 and ISC0 - 
--  Interrupt Service Control. Activate with help of the SFR, the IEN0 
--  and IEN1 enables the interrupt and the ISC0 show the status for the 
--  interrupts.	         
--------------------------------------------------------------------------	 
--	 INTERRUPT ENABLES - SFR IEN0, HEX 0C 
--     -----------------------------------------------
--    | EA  | WDT |	ET2 | ET1 | ET0 | EX2 | EX1 | EX0 |
--	  -----------------------------------------------
      elsif(WE_io  = '1' and ID_io = X"0C") then
         EA_int <= Data_in_io(7);  -- Activate or deactivate all Interrupts EA
--	   WDT_int <= Data_in_io(6);  -- Activate or deactivate WDT
--	   ET2_int <= Data_in_io(5);	-- Activate or deactivate Interrupts Timer 2
	   ET1_int <= Data_in_io(4);	-- Activate or deactivate Interrupts Timer 1
	   ET0_int <= Data_in_io(3);	-- Activate or deactivate Interrupts Timer 0
	   EX2_int <= Data_in_io(2);	-- Activate or deactivate External Interrupt 2
	   EX1_int <= Data_in_io(1);	-- Activate or deactivate External Interrupt 1
	   EX0_int <= Data_in_io(0);	-- Activate or deactivate External Interrupt 0
--	 INTERRUPT ENABLES - SFR IEN1, HEX 0D 
--     -----------------------------------------------
--    |  X  |  X  |	 X  |  X  |  X  |  X  | EC0 | ES0 |
--	  -----------------------------------------------
	 elsif(WE_io  = '1' and ID_io = X"0D") then
--	    EC0_int <= Data_in_io(1);  -- Activate or deactivate CAN-BUS Interrupt
	    ES0_int <= Data_in_io(0);	 -- Activate or deactivate Serial Interrupt

--	 INTERRUPT SERVICE CONTROL - SFR ISC0, HEX 0E 
--     -----------------------------------------------
--    | IC0 | IS0 |	IT2 | IT1 | IT0 | IX2 | IX1 | IX0 |
--	  -----------------------------------------------
      elsif(WE_io  = '1' and ID_io = X"0E") then
         IC0_int <= Data_in_io(7);	 -- Clear Interrupt FLAG for CAN-BUS 
	    IS0_int <= Data_in_io(6);	 -- Clear Interrupt FLAG for Serial 
	    IT2_int <= Data_in_io(5);	 -- Clear Interrupt FLAG for Timer 2 
	    IT1_int <= Data_in_io(4);	 -- Clear Interrupt FLAG for Timer 1
	    IT0_int <= Data_in_io(3);	 -- Clear Interrupt FLAG for Timer 0
	    IX2_int <= Data_in_io(2);	 -- Clear Interrupt FLAG for External 2
	    IX1_int <= Data_in_io(1);	 -- Clear Interrupt FLAG for External 1
	    IX0_int <= Data_in_io(0);	 -- Clear Interrupt FLAG for External 0
      elsif(RE_io  = '1' and ID_io = X"0E") then
         Data_out_io <= (IC0_int & IS0_int & IT2_int & IT1_int & IT0_int & IX2_int & IX1_int & IX0_int);
--    External Interrupt Service Routine		
 	 elsif (EX0_int='1' and IX0_int='0' and X0_int='1' and EA_int='1') then
	   IX0_int <= '1';		-- Set Interrupt FLAG
	   Interrupt_io <= '1';	-- Send Interrupt to PicoBlaze 
 	 elsif (EX1_int='1' and IX1_int='0' and X1_int='1' and EA_int='1') then
	   IX1_int <= '1';		-- Set Interrupt FLAG
	   Interrupt_io <= '1';	-- Send Interrupt to PicoBlaze  
 	 elsif (EX2_int='1' and IX2_int='0' and X2_int='1' and EA_int='1') then
	   IX2_int <= '1';		-- Set Interrupt FLAG
	   Interrupt_io <= '1';	-- Send Interrupt to PicoBlaze
------------------------------------------------------------------------------
--  This program function support the Timer option via the SFR TCON there 
--  start/stop timer to run and the reading flag function for timer interrupt. 
--  The supports also the reload value for Timer 0 and Timer 2.
------------------------------------------------------------------------------
--	 TIMER SERVICE CONTROL - SFR TCON, HEX 0F
--     -----------------------------------------------
--    |  X  |  X  |	TF2 | TR2 | TF1 | TR1 | TF0 | TR0 |
--	  -----------------------------------------------
      elsif(WE_io  = '1' and ID_io = X"0F") then  -- Write to TCON register 
	 								     -- for Start/Stop 
  	    TR0_timer <= Data_in_io(0);		     -- Start timer 0 with set a '1'
  	    TR1_timer <= Data_in_io(2);		     -- Start timer 1 with set a '1'
      elsif(RE_io  = '1' and ID_io = X"0F") then  -- Read status flag and which timer 
	 									-- there are on! 
	    Data_out_io <= ("0000" &  TF1_timer & TR1_timer & TF0_timer & TR0_timer); 

--    Timer Count 0 - TC0, HEX 10
      elsif(WE_io  = '1' and ID_io = X"10") then  -- Reload new value to timer 0
	    TC0_timer <= Data_in_io;			    
--    Timer Count 1 - TCL1, HEX 11 and TCH1 HEX 12
      elsif(WE_io  = '1' and ID_io = X"11") then  -- Reload new value to Timer 2
	    TC1_timer(7 DOWNTO 0) <= Data_in_io;     -- Set the low byte
      elsif(WE_io  = '1' and ID_io = X"12") then  
	    TC1_timer(15 DOWNTO 8) <= Data_in_io;    -- Set the high byte

--    Timer Interrupt Service Routine					  
 	 elsif (ET0_int='1' and TF0_timer='1' and IT0_int='0' and EA_int='1') then
	   IT0_int <= '1';	    -- Set Timer 0 Interrupt FLAG
	   Interrupt_io <= '1'; -- Send Interrupt to PicoBlaze	
 	 elsif (ET1_int='1' and TF1_timer='1' and IT1_int='0' and EA_int='1') then
	   IT1_int <= '1';	    -- Set Timer 1 Interrupt FLAG
	   Interrupt_io <= '1'; -- Send Interrupt to PicoBlaze

--    Relase Interrupt
	 elsif (Interrupt_ack_io = '1') then -- When Interrupt ACK is active 
	   Interrupt_io <= '0'; -- Set global Interrupt to zero	   			    

      end if;						  								  
   end process;
end in_out_ports;